/*
 * Socket driver for excelan ethernet driver
 *
 * This driver is used to:
 *	- manipulate the sockets which reside on the exos
 *
 * Written by: Kipp Hickman
 */

#ifdef SYSTEM5
#include "sys/param.h"
#include "sys/config.h"
#include "sys/types.h"
#include "sys/systm.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/tty.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/buf.h"
#endif

#ifdef VERSION7
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/dir.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/buf.h>
#include <sys/file.h>
#endif

#include <sys/extypes.h>
#include <sys/socket.h>
#include <sys/swap.h>
#include <sys/exreg.h>
#include <sys/exos.h>
#include <sys/soioctl.h>
#include <net/misc.h>
#include <net/in.h>
#include <ex_ioctl.h>
#ifndef PEGASUS
#include <sys/ioctl.h>
#endif
#include "config.h"

#define MIN(a,b)	(((a) < (b))? (a):(b))
#ifdef PEGASUS
int	ex_inited = 0;		/* Is EXOS inited? */
#endif
extern	struct msg *ex_findmsg();
struct	kludge *ex_send(), *soreceive();
struct	exctrl ex_db;			/* software device structure */

/* try to be smarter about byte swapping */
#undef	FAST

/*	#undef	DEBUG	*/
/*	#undef	DEBUG_NOISE	*/

/*
 * soopen:
 *	- called to initiate usage of a socket device
 *	- device must be free for this to work correctly, so check
 *	  the so_state and make sure it is
 *	- soclose() will clear the state
 *	- soioctl() --> SOSOCKET will set up the state
 */
/*ARGSUSED*/
soopen(dev, flag)
	dev_t dev;
	int flag;
{
	register int d = minor(dev);

#ifdef DEBUG_NOISE
	printf("soopen(dev=%d,flag=%d)\n",d,flag);
#endif
	if (so_state[d] != SOS_RESERVED)
		u.u_error = EBUSY;
}

/*
 * soclose:
 *	- called to terminate usage of a socket
 *	- send SOCLOSE message
 *	- shutdown socket usage by clearing its state and socket id
 */
/*ARGSUSED*/
soclose(dev, flag)
	dev_t dev;
	int flag;
{
	register struct msg *mp;
	register struct kludge *kp;
	register short d = minor(dev);

#ifdef DEBUG_NOISE
	printf("soclose(dev=%d,flag=%d)\n",dev,flag);
#endif
	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
	if (so_state[d] == SOS_BUSY) {
		so_state[d] = SOS_FREE;
		mp = ex_findmsg();
		mp->msg_msg.nm_soid = so_id[d];
		so_id[d] = 0;
		kp = ex_send(SOCLOSE, mp, &mp->msg_msg, 1);
		u.u_error = kp->kl_reply;
		kp->kl_state = KL_FREE;
		if (ex_db.ex_state & ST_WAITING)
			wakeup((caddr_t)&ex_db);
	} else
		so_state[d] = SOS_FREE;
}

/*
 * soread:
 *	- attempt read on a socket
 */
soread(dev)
	dev_t dev;
{
	register int d = minor(dev);
	register struct kludge *kp;

#ifdef DEBUG_NOISE
	printf("soread(dev=%d)\n",d);
#endif
	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
	if (so_state[d]) {
		if (so_dtype[d]) 
			kp = soreceive(dev, (struct sockaddr *)NULL);
		else {
			u.u_procp->p_flag |= SLOCK;
			kp = sorawrcv(dev, (struct sockaddr *)NULL);
			u.u_procp->p_flag &= ~SLOCK;
		}
		kp->kl_state = KL_FREE;
		if (ex_db.ex_state & ST_WAITING)
			wakeup((caddr_t)&ex_db);
	} else
		u.u_error = EIO;
}

sowrite(dev)
	dev_t dev;
{
	register int d = minor(dev);

#ifdef DEBUG_NOISE
	printf("sowrite(dev=%d)\n",d);
#endif
	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
	if (so_state[d]) {
		if (so_dtype[d])
			sosend(dev, (struct sockaddr *)NULL);
		else {
			u.u_procp->p_flag |= SLOCK;
			sorawsnd(dev, (struct sockaddr *)NULL);
			u.u_procp->p_flag &= ~SLOCK;
		}
	}
	else
		u.u_error = EIO;
}

/*
 * soioctl:
 *	- all the major socket system calls are emulated via the soioctl()
 *	  processor
 *	- added kflag parameter 3/8/84 so as to make soioctl calls from
 *	  the kernel ...bjc  ** removed 3/16/84  ..bjc **
 *	- added a raw sosend and soreceive that avoids kernel buffering
 *	  <if it doesn't work,then bjc did it> ...bjc
 */
/*ARGSUSED*/
soioctl(dev, cmd, addr, flag)
	dev_t dev;
	register int cmd;
	int flag;
	caddr_t addr;
{
	register struct msg *mp;
	register struct net_msg *np;
	register struct Sock_cmd *Sock_cmd;
	register int d = minor(dev);
	register struct kludge *kp;
	struct SOioctl param;

	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
#ifdef	DEBUG_NOISE
printf("unix: soioctl: dev=%x cmd=%d flag=%d\n", minor(dev), cmd, flag);
#endif
	d = minor(dev);
	if (flag & FKERNEL)
		bcopy(addr, &param, sizeof(param));
	else
		if (copyin(addr, (caddr_t)&param, sizeof(param))) {
			u.u_error = EFAULT;
			return;
		}

	switch (cmd) {
	  case SOSOCKET:
	  case SOACCEPT:
	  case SOCONNECT:
#ifdef PEGASUS
	  case SOSOCKADDR:
#else
	  case SOSOCKETADDR:
#endif
		mp = ex_findmsg();
		np = &mp->msg_msg;
		Sock_cmd = &np->nm_u.nm_cmd;
		if (Sock_cmd->nm_isaddr = param.hassa)
			Sock_cmd->nm_saddr = param.sa;
		if (Sock_cmd->nm_isproto = param.hassp)
			Sock_cmd->nm_sproto = param.sp;
		if (cmd == SOSOCKET) {
			Sock_cmd->nm_iamroot = 0;
			if (suser())
				Sock_cmd->nm_iamroot = 1;
			u.u_error = 0;
		} else
			np->nm_soid = so_id[d];
		Sock_cmd->nm_type = param.type;
		Sock_cmd->nm_options = param.options;
#ifdef DEBUG
	printf("soioctl: isaddr=%x isproto=%x iamroot=%x type=%x opt=%x\n",
		Sock_cmd->nm_isaddr,Sock_cmd->nm_isproto,Sock_cmd->nm_iamroot,
		Sock_cmd->nm_type,Sock_cmd->nm_options);
	if (Sock_cmd->nm_isaddr)  {
		printf ("soioctl: afamily=%d port=%d s_addr=%x.%x\n", Sock_cmd->nm_saddr.sin_family, htons(Sock_cmd->nm_saddr.sin_port), HIGHWORD(Sock_cmd->nm_saddr.sin_addr.s_addr), LOWWORD(Sock_cmd->nm_saddr.sin_addr.s_addr));
	}
/*	if (Sock_cmd->nm_isproto)  {
		printf ("soioctl: pfamily=%d proto=%d\n", Sock_cmd->nm_sproto.sp_family, Sock_cmd->nm_sproto.sp_protocol);
	}	*/
#endif
		kp = ex_send(cmd, mp, np, 0);
		if (u.u_error = kp->kl_reply) {
			kp->kl_state = KL_FREE;
			if (ex_db.ex_state & ST_WAITING)
				wakeup((caddr_t)&ex_db);
			return;
		}
		break;
	  case SOSEND:
	  case SORECEIVE:
		u.u_count = param.count;
		u.u_base = param.cbuf;
		if (flag & FKERNEL)
			u.u_segflg = 1;
		else
			u.u_segflg = 0;
		if (cmd == SOSEND) {
			if (so_dtype[d]) 
				sosend(dev, param.hassa ? &param.sa :
				                  (struct sockaddr *)NULL);
			else {
				u.u_procp->p_flag |= SLOCK;
				sorawsnd(dev, param.hassa ? &param.sa :
				                  (struct sockaddr *)NULL);
				u.u_procp->p_flag &= ~SLOCK;
			}
		}
		else  {
			if (so_dtype[d]) 
				kp = soreceive(dev, param.hassa ? &param.sa :
						  (struct sockaddr *)NULL);
			else {
				u.u_procp->p_flag |= SLOCK;
				kp = sorawrcv(dev, param.hassa ? &param.sa :
						  (struct sockaddr *)NULL);
				u.u_procp->p_flag &= ~SLOCK;
			}
		}
		param.count -= u.u_count;
		break;
	  case SORAWSOCK:
		so_dtype[d] = param.options;
		return;
	  default:
		if ((cmd & 0xff00) == ('s'<<8))
			socontrol(cmd, addr, so_id[d]);
		else
			u.u_error = ENXIO;
		return;
	}

    /* handle response to command */

	switch (cmd) {
	  case SOSOCKET:
		so_id[d] = kp->kl_soid;
		so_state[d] = SOS_BUSY;
		so_dtype[d] = SOREG;
		break;
	  case SOACCEPT:
	  case SOCONNECT:
#ifdef PEGASUS
	  case SOSOCKADDR:
#else
	  case SOSOCKETADDR:
#endif
	  case SORECEIVE:
	  case SOSEND:
		if (param.hassa) {
/* not portable
			param.sa = kp->kl_sa;
*/
			bcopy(&kp->kl_sa,&param.sa,sizeof(struct sockaddr));
			if (flag & FKERNEL)
				bcopy(&param, addr, sizeof(param));
			else if (copyout((caddr_t)&param, addr, sizeof(param))){
				u.u_error = EFAULT;
				break;
			}
		}
	}
	if (cmd != SOSEND)  {
		kp->kl_state = KL_FREE;		/* free up kludge buf */
		if (ex_db.ex_state & ST_WAITING)
			wakeup((caddr_t)&ex_db);
	}
}

/*
 * sosend:
 *	- write/send data to a socket
 */
sosend(dev, sa)
	dev_t dev;
	register struct sockaddr *sa;
{
	register struct Sock_pkt *Sock_pkt;
	register int amount;
	register caddr_t baddr;
	register struct msg *mp;
	register struct net_msg *np;
	register struct kludge *kp;
	register long sopaddr;
	register short d;
	struct buf *bp;

#ifdef DEBUG_NOISE
	printf("sosend(dev=%d)\n",minor(dev));
#endif
	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
	d = so_id[minor(dev)];
	bp = geteblk();
	baddr = bp->b_un.b_addr;
	sopaddr = ((long)vtop(baddr)) & 0xffffff;
	while (u.u_count) {
		amount = MIN(BSIZE, u.u_count);
		iomove(baddr, amount, B_WRITE);
		if (u.u_error)
			break;
		mp = ex_findmsg();
		np = &mp->msg_msg;
		np->nm_soid = d;
		Sock_pkt = &np->nm_u.nm_packet;
		Sock_pkt->nm_count = amount;
		if (sa) {
			Sock_pkt->nm_isaddr = 1;
			/*Sock_pkt->nm_saddr = *sa;	/* NON PORTABLE */
			bcopy(sa, &Sock_pkt->nm_saddr, sizeof(struct sockaddr));
		} else
			Sock_pkt->nm_isaddr = 0;
		Sock_pkt->nm_bufaddr = sopaddr;
		kp = ex_send(SOSEND, mp, np, 0);
		kp->kl_state = KL_FREE;
		if (ex_db.ex_state & ST_WAITING)
			wakeup((caddr_t)&ex_db);
		if (u.u_error = kp->kl_reply) {
			if (u.u_error == EPIPE)
				psignal(u.u_procp, SIGPIPE);
			break;
		}
	}
	brelse(bp);
}

/*
 * soreceive:
 *	- read/receive data from a socket
 *	- only do a single read, returning the number of bytes read
 */
struct kludge *
soreceive(dev, sa)
	dev_t dev;
	register struct sockaddr *sa;
{
	register struct Sock_pkt *Sock_pkt;
	register int amount;
	register caddr_t baddr;
	register struct msg *mp;
	register struct net_msg *np;
	register struct kludge *kp;
	register long sopaddr;
	struct buf *bp;

#ifdef DEBUG_NOISE
	printf("soreceive(dev=%d)\n",minor(dev));
#endif
	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
	bp = geteblk();
	baddr = bp->b_un.b_addr;
	sopaddr = ((long)vtop(baddr)) & 0xffffff;
	amount = MIN(BSIZE, u.u_count);

	mp = ex_findmsg();
	np = &mp->msg_msg;
	np->nm_soid = so_id[minor(dev)];
	Sock_pkt = &np->nm_u.nm_packet;
	Sock_pkt->nm_count = amount;
	if (sa) {
		Sock_pkt->nm_isaddr = 1;
		/*Sock_pkt->nm_saddr = *sa;	/* NON PORTABLE */
		bcopy(sa, &Sock_pkt->nm_saddr, sizeof(struct sockaddr));
	} else
		Sock_pkt->nm_isaddr = 0;
	Sock_pkt->nm_bufaddr = sopaddr;
	kp = ex_send(SORECEIVE, mp, np, 0);
	if (u.u_error = kp->kl_reply) {
		amount = 0;
		goto out;
	}
	amount = kp->kl_count;
	if (amount) {
		iomove(baddr, amount, B_READ);
	}

out:
	brelse(bp);
	return kp;
}

/*
 * sorawsnd:
 *	- write/send data to a socket
 *	- avoids using a temp kernel buffer. writes data directly into
 *        user's defined buffer.
 */
sorawsnd(dev, sa)
	dev_t dev;
	register struct sockaddr *sa;
{
	register struct Sock_pkt *Sock_pkt;
	register int amount;
	register caddr_t baddr;
	register struct msg *mp;
	register struct net_msg *np;
	register struct kludge *kp;
	register long sopaddr;
	register short d;
	struct buf *bp;
	unsigned base;
	int	nb;

	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
	base = u.u_base;
	nb = btop(base);
	/*
	 * Make sure user buffer isn't in text space
	 */
	 if (nb < (u.u_sep ? 0 : u.u_tsize))
		goto bad;
	/*
	 * Make sure user buffer lies within one segment
	 */
	 if ( ((long)base + (long)u.u_count) > (long)dscrlimit(UDS_SEL))
		goto bad;

	d = so_id[minor(dev)];
	sopaddr = dscraddr(UDS_SEL) + base;
printf("RAWSND u.u_count = %d\n",u.u_count);
	while (u.u_count) {
		amount = MIN(BSIZE, u.u_count);
		u.u_count -= amount;
printf("RAWSND u.u_count = %d\n",u.u_count);
		u.u_base += amount;
		u.u_offsett += amount;
		mp = ex_findmsg();
		np = &mp->msg_msg;
		np->nm_soid = d;
		Sock_pkt = &np->nm_u.nm_packet;
		Sock_pkt->nm_count = amount;
		if (sa) {
			Sock_pkt->nm_isaddr = 1;
			/*Sock_pkt->nm_saddr = *sa;	/* NON PORTABLE */
			bcopy(sa, &Sock_pkt->nm_saddr, sizeof(struct sockaddr));
		} else
			Sock_pkt->nm_isaddr = 0;
		Sock_pkt->nm_bufaddr = sopaddr;
		kp = ex_send(SOSEND, mp, np, 0);
printf("RAWSND WROTE %d\n",kp->kl_count);
		kp->kl_state = KL_FREE;
		if (ex_db.ex_state & ST_WAITING)
			wakeup((caddr_t)&ex_db);
		if (u.u_error = kp->kl_reply) {
			if (u.u_error == EPIPE)
				psignal(u.u_procp, SIGPIPE);
			break;
		}
	}
bad:
	u.u_error = EFAULT;
}

/*
 * sorawrcv
 *	- read/receive data from a socket
 *	- only do a single read, returning the number of bytes read
 *	- avoids using a temp kernel buffer. writes data directly into
 *        users defined buffer.
 */
struct kludge *
sorawrcv(dev, sa)
	dev_t dev;
	register struct sockaddr *sa;
{
	register struct Sock_pkt *Sock_pkt;
	register int amount;
	register caddr_t baddr;
	register struct msg *mp;
	register struct net_msg *np;
	register struct kludge *kp;
	register long sopaddr;
	struct buf *bp;
	unsigned base;
	int	nb;

	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
	base = u.u_base;
	nb = btop(base);
	/*
	 * Make sure user buffer isn't in text space
	 */
	 if (nb < (u.u_sep ? 0 : u.u_tsize))
		goto bad;
	/*
	 * Make sure user buffer lies within one segment
	 */
	 if ( ((long)base + (long)u.u_count) > (long)dscrlimit(UDS_SEL))
		goto bad;

	sopaddr = dscraddr(UDS_SEL) + base;
	amount =  u.u_count;

	mp = ex_findmsg();
	np = &mp->msg_msg;
	np->nm_soid = so_id[minor(dev)];
	Sock_pkt = &np->nm_u.nm_packet;
	Sock_pkt->nm_count = amount;
	if (sa) {
		Sock_pkt->nm_isaddr = 1;
		/*Sock_pkt->nm_saddr = *sa;	/* NON PORTABLE */
		bcopy(sa, &Sock_pkt->nm_saddr, sizeof(struct sockaddr));
	} else
		Sock_pkt->nm_isaddr = 0;
	Sock_pkt->nm_bufaddr = sopaddr;
	kp = ex_send(SORECEIVE, mp, np, 0);
	if (u.u_error = kp->kl_reply) {
		amount = 0;
		goto out;
	}
	u.u_count -= kp->kl_count;

out:
	return kp;
bad:
	u.u_error = EFAULT;
}

/*
 * socontrol:
 *	- process socket ioctl's (real socket ioctl's)
 */
socontrol(cmd, addr, soid)
	int cmd;
	caddr_t addr;
	int soid;
{
	register struct msg *mp;
	register struct net_msg *np;
	register struct kludge *kp;
	register struct Sock_ioctl *sp;
	register int howmuch;		/* 0==int, 1==char */
	int	anint;
	char	achar;

#ifdef DEBUG_NOISE
	printf("socontrol(cmd=%d,soid=%d)\n",cmd,soid);
#endif
	if (!ex_inited) {
		u.u_error = ENXIO;
		return;
	}
    /* figure out which ioctl we are doing ... */

	switch (cmd) {
#ifndef PEGASUS
	  case FIONREAD:
#endif
	  case SIOCGKEEP:
	  case SIOCGLINGER:
	  case SIOCRCVOOB:
	  case SIOCATMARK:
		break;
	  case FIONBIO:
	  case SIOCSKEEP:
	  case SIOCSLINGER:
		if (copyin(addr, (caddr_t)&anint, sizeof(anint))) {
			u.u_error = EFAULT;
			return;
		}
		howmuch = 0;
		break;
	  case SIOCSENDOOB:
		if (copyin(addr, (caddr_t)&achar, sizeof(achar))) {
			u.u_error = EFAULT;
			return;
		}
		howmuch = 1;
		break;
	  case SIOCSPGRP:
	  case SIOCGPGRP:
		return;
	  default:
		u.u_error = EOPNOTSUPP;
		return;
	}

    /* send command to board */

	mp = ex_findmsg();
	np = &mp->msg_msg;
	np->nm_soid = soid;
	sp = &np->nm_u.nm_ioctl;
	sp->nm_ioccmd = cmd;
	switch (howmuch) {
	  case 0:
		sp->nm_long = anint;		/* not portable */
		break;
	  case 1:
		sp->nm_long = achar;		/* not portable */
		break;
	}
	kp = ex_send(SOIOCTL, mp, np, 0);
	if (u.u_error = kp->kl_reply) {
		kp->kl_state = KL_FREE;
		if (ex_db.ex_state & ST_WAITING)
			wakeup((caddr_t)&ex_db);
		return;
	}

    /* examine reply (if needed) */

	switch (cmd) {
#ifndef PEGASUS
	  case FIONREAD:
		if (copyout((caddr_t)&kp->kl_along, addr, sizeof(kp->kl_along)))
			u.u_error = EFAULT;
		break;
#endif
	  case SIOCGKEEP:
	  case SIOCGLINGER:
	  case SIOCATMARK:
		if (copyout((caddr_t)&kp->kl_anint, addr, sizeof(kp->kl_anint)))
			u.u_error = EFAULT;
		break;
	  case SIOCRCVOOB:
		if (copyout((caddr_t)&kp->kl_achar, addr, sizeof(kp->kl_achar)))
			u.u_error = EFAULT;
		break;
	}
	kp->kl_state = KL_FREE;
}
