/*
 *	VIRTUAL DISK
 */

/*
 * Insure DEBUG for now.
 */
#ifndef	DEBUG
#define	DEBUG	1
#endif

#ifdef	DEBUG
int	debugvd = 1;			/* 0=off,1=min,2=max */
#endif

#include	"../h/param.h"
#include	"../h/systm.h"
#include	"../h/buf.h"
#include	"../h/conf.h"
#include	"../h/dir.h"
#include	"../h/user.h"
#include	"../h/file.h"
#include	<sys/socket.h>
#include	<sys/extypes.h>
#include	<net/misc.h>
#include	<net/in.h>
#include	<ex_ioctl.h>
#include	"vd.h"

struct	buf	vdtab;
struct	buf	rvdbuf;
struct	buf	cvdbuf;		/* command buffer, used for `q' */

struct	vd_device	vd_dev;

Uchar	vd_state[NVDISKS];

/*
 *	vdinit
 */

vdinit ()
{
	register	int	err_sav;
	register	struct	vd_device	*dp;
	dev_t	sock_dev;
	struct	sockad_in *sin;
	struct	SOioctl	VDioctl;

#ifdef	DEBUG
	if (debugvd)
		printf ("vdinit called\n");
#endif
	dp = &vd_dev;
	exioctl (0, EXIOCFIND, &sock_dev, FKERNEL);
	if (u.u_error)
		return;
	dp->vdskt = sock_dev;
	soopen (sock_dev, 2);
	if (u.u_error)  {
		soclose (sock_dev, 0);
		return;
	}
	VDioctl.hassa = 1;
	sin = &VDioctl.sa;
	sin->sin_family = AF_INET;
	sin->sin_port = htons (IPPO_VDD);
	sin->sin_addr.s_addr = 0;
	VDioctl.hassp = 0;
	VDioctl.type = SOCK_STREAM;
	VDioctl.options = SO_KEEPALIVE;
	soioctl (sock_dev, SOSOCKET, &VDioctl, FKERNEL);
	if (err_sav = u.u_error)  {
		soclose (sock_dev, 0);
		u.u_error = err_sav;
		return;
	}

	sin->sin_addr.s_addr = VAXADDR;
	sin->sin_port = htons (VDPORT);
	soioctl (sock_dev, SOCONNECT, &VDioctl, FKERNEL);
	if (err_sav = u.u_error)  {
		soclose (sock_dev, 0);
		u.u_error = err_sav;
		return;
	}

	++dp->vdinited;
}

/*
 * vdopen
 */

vdopen(dev, flag)
register dev_t dev;
{
	register	int	d = minor(dev);
	register	char	*cp1;       
	register	struct	vd_device	*dp;

#ifdef	DEBUG
	if (debugvd)
		printf("vdopen called: dev=%x\n", d);
#endif
reinit:
	dp = &vd_dev;
	if (!dp->vdinited)  {
		vdinit (d);
		if (u.u_error)
			return;
	}
	if (vd_state[d]) {
		u.u_error = ENXIO;
		return;
	}
	dp->vdblk = 0;
	dp->vdbc = 1;
	dp->vddv = d;
	dp->vdcsr = CMD_OP|GO;
	vdio ();
	if (dp->vdcsr & ERR)  {
		soclose (dp->vdskt, 0);
		dp->vdinited = 0;
		u.u_error = 0;
		goto reinit;
	}
	vd_state[d] = 1;
	return;
}

/*
 * vdclose
 * 	Send the sign-off message
 * 	allow further opens
 */

vdclose(dev)
register dev_t dev;
{
	register	int	d = minor(dev);
	register struct buf *bp;

#ifdef	DEBUG
	if (debugvd)
		printf("vdclose called: dev=%x\n", d);
#endif
	if (!vd_state[d])  {
		u.u_error = ENXIO;
		return;
	}
/*
	if ((d & 0100) == 0) {
		bp = &cvdbuf;
		spl6();
		while (bp->b_flags&B_BUSY) {
			bp->b_flags |= B_WANTED;
			sleep((caddr_t)bp, PRIBIO+1);
		}
		spl0();
		bp->b_flags = B_BUSY | B_PHYS;
		bp->b_dev = dev;
		bp->b_resid = CMD_QT;
		bp->b_bcount = 1;
		vdstrategy(bp);
		spl6();
		while ((bp->b_flags&B_DONE) == 0) {
			sleep((caddr_t)bp, PRIBIO+1);
		}
		bp->b_flags &= ~B_BUSY;
		if (bp->b_flags & B_WANTED)
			wakeup((caddr_t) bp);
	}
	spl0();
*/
	vd_state[d] = 0;
}

vdstrategy(bp)
register struct buf *bp;
{
#ifdef	DEBUG
	if (debugvd) {
		printf("vdstrat called: bp=%x,blk=%D,cnt=%d,%s\n",
			bp, bp->b_blkno, bp->b_bcount,
			(bp->b_flags&B_READ)?"READ":"WRITE");
	}
#endif
	if (bp->b_blkno >= NVDBLKS) {
		bp->b_flags |= B_ERROR;
		iodone(bp);
		return;
	}
	bp->av_forw = 0;
	spl5();
	if (vdtab.b_actf == NULL)
		vdtab.b_actf = bp;
	else
		vdtab.b_actl->av_forw = bp;
	vdtab.b_actl = bp;
	if (vdtab.b_active == NULL)
		vdstart();
	spl0();
}

/*
 * vdintr
 *	"interrupt" has occured -- end of transfer.
 */

vdintr()
{
	register struct buf *bp;
	register c;

#ifdef	DEBUG
	if (debugvd)
		printf("vdintr called\n");
#endif
	if (vdtab.b_active == NULL)
		printf ("VDINTR: vdtab.b_active is NULL\n");
	c = vd_dev.vdcsr;
	if((c & DONE) == 0) {
		printf ("VDINTR: vd_dev.vdcsr is not DONE\n");
	}
	bp = vdtab.b_actf;
	vdtab.b_active = NULL;
	if (c & ERR) {
		++vdtab.b_errcnt;
		deverror(bp, c, vd_dev.vdcsr);
/*
		if (++vdtab.b_errcnt <= 10) {
			vdstart();
		}
		bp->b_flags |= B_ERROR;
		return;
*/
		printf ("VDINTR: DEVERROR\n");
	}
	else
		vdtab.b_errcnt = 0;
	vdtab.b_actf = bp->av_forw;
	bp->b_resid = 0;
	iodone(bp);
}

vdread (dev)
dev_t	dev;
{
	register	int	d = minor(dev);

#ifdef	DEBUG
	if (debugvd)
		printf("vdread called: dev=%x\n", d);
#endif
	physio(vdstrategy, &rvdbuf, dev, B_READ);
}

vdwrite (dev)
dev_t	dev;
{
	register	int	d = minor(dev);

#ifdef	DEBUG
	if (debugvd)
		printf("vdwrite called: dev=%x\n", d);
#endif
	physio(vdstrategy, &rvdbuf, dev, B_WRITE);
}

vdioctl (dev, cmd, addr, flag)
dev_t	dev;
int	cmd, flag;
caddr_t	addr;
{
#ifdef	DEBUG
	if (debugvd)
		printf ("vdioctl called: (dev=%d,cmd=%d)\n", minor(dev), cmd);
#endif
}

vdstart ()
{
	register struct buf *bp;
	register struct vd_device *dp;

#ifdef	DEBUG
	if (debugvd)
		printf("vdstart called\n");
#endif
next:
	if ((bp = vdtab.b_actf) == NULL)
		return;
	vdtab.b_active++;
	dp = &vd_dev;
	dp->vddv = minor (bp->b_dev);
	dp->vdblk = (int) bp->b_blkno;
	dp->vdbc = bp->b_bcount;
	dp->vdaddr = (((long) bp->b_xmem)<<16) | (unsigned) bp->b_un.b_addr;
	dp->vdcsr = (bp->b_flags&B_READ? CMD_RD:CMD_WR) | GO;
	if (bp == &cvdbuf)
		dp->vdcsr = bp->b_resid|GO;
	vdio();
	vdintr();
	goto next;
}



/*	vdio - Emulate the Pseudo Disk
 *
 *	vdio is called to emulate the action of the pseudo disk hardware.
 *	The command is setup into a set of 'device registers' as if for
 *	a real device.  vdio is then called to fake the I/O.
 *
 *      vdio builds and sends the command buffer to the host machine,
 *      then, for each integral block of at most BUFFSIZE bytes,
 *
 *      If Read,
 *              read and acknowledge buffers from the host until satisfied, or
 *
 *      If Write,
 *              break data into pieces, and transmit each piece.
 */


vdio()
{
	struct	sockad_in	*sin;
	register	struct	vd_device	*dp;
	register	char	*cp1;
	register	int	i, dev;
	register	int	cmd;
	static	int	active = 0;
	struct	SOioctl	VDioctl;
	struct  {
		caddr_t x_base;
		unsigned int x_count;
		off_t   x_offset;
	}  save;
	char	vddata[MESGSIZE];

	
#ifdef	DEBUG
	if (debugvd)
		printf("vdio called\n");
#endif
	if (active++)
		printf ("VDIO: multiple call of vdio\n");

	dp = &vd_dev;                           
	if ((dp->vdcsr & GO) == 0)
		printf ("VDIO: vd_dev.vdcsr not set to GO\n");
	dp->vdcsr &= ~(GO | ERR);               
	dev = dp->vdskt;
	save.x_base = u.u_base;
	save.x_count = u.u_count;
	save.x_offset = u.u_offset;

	VDioctl.hassa = 0;
	cmd = dp->vdcsr & CMD;

	while (dp->vdbc > 0) {
		cp1 = &vddata[0];
		switch (cmd) {
			case CMD_WR:      
				*cp1++ = WCOM;		
				break;
			case CMD_RD:    
				*cp1++ = RCOM;		
				break;
			case CMD_OP:      
				*cp1++ = OCOM;		
				break;
			default:
#ifdef	DEBUG
			printf ("VDIO: bad command code %d\n", dp->vdcsr);
#endif
				dp->vdcsr |= ERR;
				goto bad;
		}
		*cp1++ = ' ';
		*cp1++ = dp->vddv;		
		*cp1++ = ' ';
		*cp1++ = dp->vdblk;         
		dp->vdblk++;
		*cp1++ = 0;

		VDioctl.cbuf = vddata;
		VDioctl.count = MESGSIZE;	
		soioctl (dev, SOSEND, &VDioctl, FKERNEL);
		if (u.u_error)  {
			dp->vdcsr |= ERR;
			goto bad;
		}
		VDioctl.cbuf = vddata;
		VDioctl.count = MESGSIZE;	
		soioctl (dev, SORECEIVE, &VDioctl, FKERNEL);
		if (u.u_error)  {
			dp->vdcsr |= ERR;
			goto bad;
		}
		cp1 = &vddata[0];
		if (*cp1 != SUCC) {             
			dp->vdcsr |= ERR;
#ifdef	DEBUG
	printf ("VDIO: %d command returned ERR: %s\n", cmd, cp1);
#endif
			goto bad;
		}
		switch (cmd) {
			case CMD_WR:      
				u.u_base = save.x_base;
				u.u_count = MIN (BSIZE, save.x_count);
				u.u_offset = save.x_offset;
				u.u_segflg = 0;
				sowrite (dev);
				if (u.u_error)  {
					dp->vdcsr |= ERR;
					goto bad;
				}
				VDioctl.cbuf = vddata;
				VDioctl.count = MESGSIZE;	
				soioctl (dev, SORECEIVE, &VDioctl, FKERNEL);
				if (u.u_error)  {
					dp->vdcsr |= ERR;
					goto bad;
				}

				cp1 = &vddata[0];
				if (*cp1 != SUCC) {             
					dp->vdcsr |= ERR;
#ifdef	DEBUG
	printf ("VDIO: WRITE command returned ERR: %s\n", cp1);
#endif
					goto bad;
				}
				break;

			case CMD_RD:    
				cp1 = &vddata[0];
				*cp1++ = RDYCOM;		
				*cp1++ = 0;
				VDioctl.cbuf = vddata;
				VDioctl.count = MESGSIZE;	
				soioctl (dev, SOSEND, &VDioctl, FKERNEL);
				if (u.u_error)  {
					dp->vdcsr |= ERR;
					goto bad;
				}
	
				u.u_base = save.x_base;
				u.u_count = MIN (BSIZE, save.x_count);
				u.u_offset = save.x_offset;
				u.u_segflg = 0;
				soread (dev);
				if (u.u_error)  {
					dp->vdcsr |= ERR;
					goto bad;
				}
				break;

			case CMD_OP:      
				goto out;
		}
		dp->vdbc -= BUFFSIZE;
		dp->vdaddr += BUFFSIZE;
		save.x_base += BUFFSIZE;
	}

out:
	dp->vdcsr = (dp->vdcsr&~(BUSY)) | (DONE);
bad:
	active = 0;                     
}
